/*
 * hwreset.c
 *
 * This tool power cycles (or powers off) the IPMI system.
 *    
 * Author:  Andy Cress  andrew.r.cress@intel.com
 * Copyright (c) 2002 Intel Corporation. 
 *
 * 01/08/02 Andy Cress - created
 * 01/31/02 Andy Cress - converted to use ipmi_cmd_ia, 
 *                       added more user options
 * 02/06/02 Andy Cress - added ipmi_cmd_va
 * 02/22/02 Andy Cress - added -s to reboot to service partition
 * 07/02/02 Andy Cress v1.3 added more Usage text
 * 08/02/02 Andy Cress v1.4 moved common ipmi_cmd() code to ipmicmd.c
 * 09/24/02 Andy Cress - stubbed in OS shutdown option
 * 01/29/03 Andy Cress v1.5 added MV OpenIPMI support
 * 04/08/03 Andy Cress v1.6 added OS shutdown option (-o)
 * 05/02/03 Andy Cress v1.7 leave console redir alone in SET_BOOT_OPTIONS
 * 05/05/04 Andy Cress v1.8 call ipmi_close before exit, did WIN32 port.
 * 08/09/04 Andy Cress v1.9 make sure to show error if ccode != 0, and
 *                          detect Langley platforms to do special 
 *                          watchdog method for -o option.
 * 11/01/04 Andy Cress 1.10 add -N / -R for remote nodes   
 * 11/16/04 Andy Cress 1.11 add -U for remote username
 * 11/30/04 Andy Cress 1.12 fix bug 1075550 with -o -N, skip -o if not local.
 */
/*M*
Copyright (c) 2002, Intel Corporation
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met:

  a.. Redistributions of source code must retain the above copyright notice, 
      this list of conditions and the following disclaimer. 
  b.. Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation 
      and/or other materials provided with the distribution. 
  c.. Neither the name of Intel Corporation nor the names of its contributors 
      may be used to endorse or promote products derived from this software 
      without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR 
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON 
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *M*/
#ifdef WIN32
#include <windows.h>
#include <stdio.h>
#include "getopt.h"
#else
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <getopt.h>
#endif
#include <string.h>
#include "ipmicmd.h"
 
#define platTSR   1    /*Langleys (use Sahalee), use alt soft-shutdown. */
#define platMBMC  2    /*mini-BMC platforms */

/*
 * Global variables 
 */
char   fdebug    = 0;
char * progver   = "1.12";
char * progname  = "hwreset";
uchar  ipmi_maj;
uchar  ipmi_min;
char   platform  = 0;
int    shuttime  = 60;  /* shutdown timeout in seconds */

static int set_wdt(uchar val)
{
        uchar idata[6];
        uchar rdata[16];
        int rlen = 8;
        uchar ccode;
        int ret, t;
 
        t = val * 10;      /* val is in sec, make timeout in 100msec */
        if ((ipmi_maj > 1) ||      /* IPMI 1.5 or greater */
            (ipmi_maj == 1 && ipmi_min >= 5))
             idata[0] = 0x44;    /* DontLog=0, DontStop=1 & use SMS/OS */
        else idata[0] = 0x04;    /* IPMI 1.0 or less */
        idata[1] = 0x01;   /* action: no pretimeout, hard reset action */
		/* 0x02;    * action value for power down instead */
        idata[2] = 0;      /* pretimeout: 30 sec (but disabled) */
        idata[3] = 0x10;   /* clear SMS/OS when done */
        idata[4] = t & 0x00ff;   /*timeout in 100msec: 0x4B0 = 1200. */
        idata[5] = (t & 0xff00) >> 8;
        ret = ipmi_cmd(WATCHDOG_SET, idata, 6, rdata, &rlen, &ccode, fdebug);
        if (ret == 0 && ccode != 0) ret = ccode;

	if (ret == 0) {  /* Start the timer by issuing a watchdog reset */
	   ret = ipmi_cmd(WATCHDOG_RESET,idata,0, rdata,&rlen, &ccode,fdebug);
	   if (ret == 0 && ccode != 0) ret = ccode;
	}
        return(ret);
}  /*end set_wdt()*/

static int IPMI_Reset(uchar bpower, uchar bootopt)
{
	uchar responseData[MAX_BUFFER_SIZE];
	int responseLength = MAX_BUFFER_SIZE;
	uchar completionCode;
	uchar inputData[6];
	int status = 0;

        /* May need to GetSystemBootOptions first to fill these in. */
	inputData[0] = 0x05;  // param, 5 = boot flags
	inputData[1] = 0x80;  // valid flags
        if (bootopt == 1)     // set up for service partition boot 
	     inputData[2] = 0x10;  // boot to svc partition
	else inputData[2] = 0x00;  // normal boot
	inputData[3] = 0x48;  // verbose disply, bypass pswd, same console redir
	inputData[4] = 0x00;  // no overrides
	inputData[5] = 0x00;  // 
        status = ipmi_cmd(SET_BOOT_OPTIONS, inputData, 6, responseData,
                        &responseLength, &completionCode, fdebug);
	if (status != ACCESS_OK || completionCode != 0) {
            printf("set_boot_options ccode %x, resp[0] = %x, resp[1] =  %x\n",
                   completionCode, responseData[0], responseData[1]);
	   }

	if ((bpower == 5) && (platform == platTSR) && (!fipmi_lan)) {  
	   /*
	    * Special OS shutdown method for Langleys (TSRLT2, etc) 
	    * Set up a watchdog event to do reset after timeout. 
	    * Valid on other platforms too if they support watchdog.
	    * Note that the "init 0" only makes sense if local. 
	    */
	   status = set_wdt((uchar)shuttime);
	   if (status == 0) {
		status = system("init 0"); /* do the OS shutdown */
	   }
	   /* Note that this will generate a Watchdog 2 event in the SEL. */
	   return(status);
	}

        /* 0 = power down, 1 = power up, 2 = power cycle, 3 = hard reset */
        /* 4 = NMI interrupt, 5 = soft shutdown OS via ACPI  */
	if (bpower > 5) bpower = 3;  /* if invalid, use default */
	inputData[0] = bpower;  // hard reset
        status = ipmi_cmd(CHASSIS_CTL, inputData, 1, responseData,
                        &responseLength, &completionCode, fdebug);
        if (fdebug) {
                printf("Cmd %x code %x, resp[0] = %x, resp[1] =  %x\n",
		       CHASSIS_CTL, completionCode, 
		       responseData[0], responseData[1]);
                }
	if (status == ACCESS_OK && completionCode == 0) {
		//successful, done
		return(0);
	} else {
	   if (bpower == 5 && completionCode == 0xcc) {
		/* See IPMI spec 22.3 Chassis Control, Table 22-4 */
		printf("Optional soft-shutdown mode not supported\n");
		/* If get here, need to use method like platTSR above. */
	   } else
        	printf("Error %d: chassis completion code = %x\n", 
			status,completionCode); 
	   return(-1);
	   }

}  /*end IPMI_Reset()*/

#ifdef WIN32
int __cdecl
#else
int
#endif
main(int argc, char **argv)
{
   int ret;
   char c;
   uchar breset;
   uchar bopt;
   uchar fshutdown = 0;
   uchar devrec[16];

   // progname = argv[0];
   breset = 3;  /* hard reset as default */
   bopt = 0;    /* Boot Options default */
   printf("%s ver %s\n", progname,progver);

   while ( (c = getopt( argc, argv,"dconrsN:R:U:x?")) != EOF ) 
      switch(c) {
          case 'd': breset = 0;     break;  /* power down */
          // case 'u': breset = 1;  break;  /* power up */
          case 'c': breset = 2;     break;  /* power cycle */
          case 'o': fshutdown = 1;  break;  /* shutdown OS */
          case 'n': breset = 4;     break;  /* interrupt (NMI) */
          case 'r': breset = 3;     break;  /* hard reset */
          case 's': bopt   = 1;     break;  /* hard reset to svc part */
#ifndef WIN32
          case 'N': gnode  = optarg;   /* nodename */
		    fipmi_lan = 1; 
		    break;
	  case 'U':
		guser = strdup(optarg);  /*remote username */
		if (!guser)  perror("strdup");
		else {  /* Hide password from 'ps' */
	        	memset(optarg, ' ', strlen(optarg));	
		}
		break;
 	  case 'R': 
		gpswd = strdup(optarg);  /*remote password */
		if (!gpswd) perror("strdup");
		else {  /* Hide password from 'ps' */
	        	memset(optarg, ' ', strlen(optarg));	
		} 
 		break;
#endif
          case 'x': fdebug = 1;     break;  /* debug messages */
	  default:
                printf("Usage: %s [-cdinrsx -N node -R pswd]\n", progname);
                printf(" where -c  power Cycles the system\n");
                printf("       -d  powers Down the system\n");
                printf("       -r  hard Resets the system\n");
                printf("       -n  sends NMI to the system\n");
                printf("       -o  soft-shutdown OS and reset\n");
                printf("       -s  reboots to Service Partition\n");
                printf("       -x  show eXtra debug messages\n");
#ifndef WIN32
                printf("       -N node  Nodename of target system\n");
                printf("       -R pswd  Remote nodename's password\n");
		printf("       -U user  Remote nodename's username\n");
#endif
                exit(1);
      }

   /* 
    * Check the Device ID to determine the platform type.
    */
   ret = ipmi_getdeviceid(devrec,16,fdebug);
   if (ret != 0) {
        printf("Cannot do ipmi_getdeviceid, ret = %d\n",ret);
	ipmi_close();
        exit(1);
   } else {
      char *pstr;
      ushort prodid;
      uchar j;

      if (fdebug) {
	printf("devid: ");
	for (j = 0; j < 16; j++) printf("%02x ",devrec[j]);
	printf("\n");
      }
      ipmi_maj = devrec[4] & 0x0f;
      ipmi_min = devrec[4] >> 4;
      if ((devrec[6] == 0x22) && (devrec[7] == 0x03)) { /*NSC mBMC*/
	  pstr = "mBMC";
	  platform = platMBMC;
      } else {  /* Intel Sahalee = 57 01 */
	  pstr = "BMC";
	  prodid = devrec[9] + (devrec[10] << 8);
	  if (prodid == 0x001b || prodid == 0x000c)  /*Langleys*/
	        platform = platTSR;
	  else if (fdebug) printf("product id = %04x\n",prodid);
      }
      printf("-- %s version %x.%x, IPMI version %d.%d \n",
             pstr, devrec[2],  devrec[3], ipmi_maj, ipmi_min);
   }

   if (fshutdown) {
	/* 
	 * Since we are user mode, we can't let it fully shut down. 
	 * IPMI can trigger this by emulating an overtemp event.
	 * There is also an watchdog/init0 method used by platTSR. 
	 */
        breset = 5;     /* soft shutdown option */
   }
   if (breset == 0) printf("%s: powering down ...\n",progname);
   else if (breset == 5) printf("%s: shutting down ...\n",progname);
   else       printf("%s: resetting ...\n",progname);
   ret = IPMI_Reset(breset,bopt);

   if (ret == 0) {   /* if ok */
	printf("%s: IPMI_Reset ok\n",progname);
	/* It usually starts resetting by this point anyway. */
   } else {
	printf("%s: IPMI_Reset error %d\n",progname,ret);
	ipmi_close(); 
   }
   exit(ret);
}  /* end main()*/

/* end hwreset.c */
