/*M*
//  PVCS:
//      $Workfile:   ipmignu.c  $
//      $Revision:   1.0  $
//      $Modtime:   21 Oct 2004 15:31:14  $
//      $Author:   andrew.r.cress@intel.com  $  
//
//  This implements support for the GNU FreeIPMI interface.
//  The latest FreeIPMI source is at http://www.gnu.org/software/freeipmi/.
//  See also ../freeipmi/README.
// 
//  10/21/03 ARC - created.
//  10/26/04 ARC - kcs path working
//  10/29/04 ARC - lan path working
//  11/10/04 ARC - use MD5 instead of MD2 for ipmi_lan,
//		   changes for comp code in _lan_cmd & _kcs_cmd,
//                 added sig_timeout routine
//  11/11/04 ARC - added DEBUG_LIB flag
//  11/23/04 ARC - added ipmi_lan ping before open_session, 
//		   added connect_state
 *M*/
/*----------------------------------------------------------------------*
The BSD License 

Copyright (c) 2002, Intel Corporation
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met:

  a.. Redistributions of source code must retain the above copyright notice, 
      this list of conditions and the following disclaimer. 
  b.. Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation 
      and/or other materials provided with the distribution. 
  c.. Neither the name of Intel Corporation nor the names of its contributors 
      may be used to endorse or promote products derived from this software 
      without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR 
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON 
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *----------------------------------------------------------------------*/

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <netdb.h>
#include <signal.h>

#include <freeipmi/freeipmi.h>

#define uchar      unsigned char
#define MAXHOSTNAMELEN    64
#define MAX_LEN_RQ   	25
#define MAX_LEN_RS   	72
// #define DEBUG_LIB   1

extern FILE *fperr;  /*defined in ipmicmd.c*/
#ifdef __IA64__
int kcs_port = 0x8a2;  
#else
int kcs_port = 0xca2;  /* KCS:  use 0xca2 for ia32, 0x8a2 for ia64 */
                       /*       some use 0xca8/0xcac w register spacing */
		       /* SMIC: use 0xca9 & 0x8a9 */
#endif
int polltime = IPMI_KCS_SLEEP_USECS;  /*=0x01*/
int gnu_retry_wait_time = 1;
int gnu_retry_attempts  = 5;
/* 
 * These variables pertain to ipmi_lan, for the node given at open time.
 * The assumption here is that these are utilities, so no more than one
 * node will be open at a given time.  
 * See also gnode, guser, gpswd in ipmicmd.c
 */
static char nodename[33] = "";
static int sockfd = 0;
static char _dest[MAXHOSTNAMELEN+1];
static char _dest_ip[INET_ADDRSTRLEN+1];
static int  _destaddr_len = 0;
static struct sockaddr_in _destaddr;
static struct sockaddr_in _srcaddr;
static uchar 	auth_type = 0;
static char *authcode = NULL;
static int authcode_len = 0;
static u_int32_t session_id = 0;
static u_int32_t session_seqnum = 0x01;   /* inbound */
static u_int32_t rq_seq = 1;		  /* outbound */
static int ipmi_timeout = 10;  /* 10 sec */
static int connect_state = 0;   
	/*  0 = init, 1 = socket() complete, 2 = bind/gethost complete,
	 *  3 = ping sent, 4 = pong received, 5 = session activated.  */
static char *conn_state_str[6] = {
	"init state", "socket complete", "gethost complete", 
	"ping sent", "pong received", "session activated" };

#ifdef DEBUG_LIB 
extern int fdebug_ipmilan;
#endif

#define GERR_TIMEOUT  -4
#define GERR_ABORT    -3
#define GERR_CONNECT  -2  /*problem connecting to BMC*/
#define GERR_OTHER    -1

void
sig_timeout(int sig)
{
   alarm(0);
   signal(SIGALRM,SIG_DFL);
   printf("ipmignu_cmd timeout, after %s\n",conn_state_str[connect_state]);
   exit(GERR_TIMEOUT);
}

void
sig_abort(int sig)
{
   static int sig_aborting = 0;
   uchar buf_rs[4]; 
   fiid_obj_t cmd_rs;
   int rv;

   if (sig_aborting == 0) {
     sig_aborting = 1;
     if (sockfd != 0) {  /* socket is open */
	  if (session_id != 0) {  /* session is open */
  	    cmd_rs = buf_rs;
	    rv = ipmi_lan_close_session(sockfd, (struct sockaddr *)&_destaddr, 
			  _destaddr_len, auth_type, session_seqnum, session_id,
			  authcode, authcode_len, rq_seq, session_id, cmd_rs);
	  }
	  close(sockfd);
     }
     signal(SIGINT,SIG_DFL);
     printf("ipmignu_cmd interrupt, after %s\n", conn_state_str[connect_state]);
     exit(GERR_ABORT);
   }  /*endif*/
} /*end sig_abort*/


static int nodeislocal(char *nodename)
{
   if (nodename == NULL) return 1;
   if (nodename[0] == 0) return 1;
   if (strcmp(nodename,"localhost") == 0) return 1;
   return 0;
}

/* signal handlers + sleep(3) is a bad idea */
static int 
_sleep(unsigned int sleep_len)
{
  struct timeval tv;
  
  if (sleep_len == 0)
    return 0;

  tv.tv_sec = sleep_len;
  tv.tv_usec = 0;
  if (select(1, NULL, NULL, NULL, &tv) < 0)
    {
      if (errno != EINTR) return(errno);
    }
  return 0;
}

static int
_kcs_cmd(char *str, int retry_wait_time, int retry_attempt, u_int8_t netfn,
     fiid_obj_t cmd_rq, fiid_template_t tmpl_rq,
     fiid_obj_t cmd_rs, fiid_template_t tmpl_rs, int fdebugcmd)
{
  u_int64_t comp_code;
  int retry_count = 0;
  int rv = 0;

  cmd_rs[1] = 0;  /*initialize completion code*/
  if (str == NULL || retry_wait_time < 0 || retry_attempt < 0 ||
      cmd_rq == NULL || tmpl_rq == NULL || cmd_rs == NULL || tmpl_rs == NULL)
  	return(-3);;

  while (1)
  {
      alarm(ipmi_timeout);
      rv = ipmi_kcs_cmd_interruptible(kcs_port, IPMI_BMC_IPMB_LUN_BMC, netfn, 
				cmd_rq, tmpl_rq, cmd_rs, tmpl_rs);
      alarm(0);
      if (rv < 0) {
          if (errno != EAGAIN && errno != EBUSY) {
	      if (fdebugcmd)
		  fprintf(fperr,"%s: ipmi_kcs_cmd_interruptible: %s \n", 
			str, strerror(errno));
              return (GERR_CONNECT);
          } else {
              if (retry_count >= retry_attempt) {
	 	if (fdebugcmd)
                  fprintf(fperr,"%s: ipmi_kcs_cmd_interruptible: BMC too busy: "
                          "retry_wait_time=%d, retry_attempt=%d", 
                          str, retry_wait_time, retry_attempt);
                return (GERR_CONNECT);
              }
              rv = _sleep(retry_wait_time);
              retry_count++;
            }
      } else { /* success, done */
	  break;
      }
  }

  /* could get cc directly from cmd_rs[1] or use IPMI_COMP_CODE(cmd_rs) */
  // could use _FIID_OBJ_GET(cmd_rs, tmpl_rs, "comp_code", &comp_code, str);
  comp_code = IPMI_COMP_CODE(cmd_rs);

  if (rv == 0 && comp_code != IPMI_COMMAND_SUCCESS) {
	if (fdebugcmd)
	   fprintf(fperr,"%s: cmd error: 0x%x\n", str, comp_code);
	rv = comp_code;
  } 

  return (rv);
}  /*end _kcs_cmd()*/


static int
_lan_cmd(char *str, struct sockaddr *hostaddr, int hostaddr_len, u_int8_t netfn,
     fiid_obj_t cmd_rq, fiid_template_t tmpl_rq,
     fiid_obj_t cmd_rs, fiid_template_t tmpl_rs, int fdebugcmd)
{
  int rv = 0;
  uchar cc;

  /* Note: str = nodename.  Could be used if more than one node. */

  memset(&cmd_rs[1],0,4);  /*init minimal recv data*/
  if (fdebugcmd) fprintf(fperr,"_lan_cmd[%x]: rq_seq = %d\n",
			 cmd_rq[0],rq_seq);
  alarm(ipmi_timeout);
  rv = ipmi_lan_cmd(sockfd, hostaddr, hostaddr_len, 
		    auth_type, session_seqnum, session_id,  
		    authcode, authcode_len, 
		    netfn, IPMI_BMC_IPMB_LUN_BMC, rq_seq,
		    cmd_rq, tmpl_rq, cmd_rs, tmpl_rs);
  alarm(0);
  cc = cmd_rs[1];
  if (fdebugcmd) fprintf(fperr,"_lan_cmd[%x]: rv = %d, cc=%x, errno = %d\n",
			 cmd_rq[0],rv,cc,errno);
  if (rv == 0 && cc != 0) {     /* completion code error */
	rv = cc;
  	if (fdebugcmd) {
		fprintf(fperr,"cmd_rq: %02x %02x %02x %02x \n",
			cmd_rq[0],cmd_rq[1], cmd_rq[2],cmd_rq[3]);
		fprintf(fperr,"cmd_rs: %02x %02x %02x %02x \n",
			cmd_rs[0],cmd_rs[1],cmd_rs[2],cmd_rs[3]);
	}
  }

  {    /* increment seqnum - even if error */
	rq_seq++;
        rq_seq = (rq_seq % (IPMI_LAN_SEQ_NUM_MAX + 1));
	session_seqnum++;
  }
  
  return (rv);
}  /*end _lan_cmd()*/

int ipmi_open_gnu(char *node, char *user, char *pswd, int fdebuglan)
{
   char *username;
   uchar priv_level;
   int rv = -1;
   uchar asf_pkt[40] = {06,0,0xFF,06,0xBE,0x11,0,0,0x80,1,0,0 };
   struct sockaddr from_addr;
   int from_len;

#ifdef DEBUG_LIB
   if (fdebuglan) fdebug_ipmilan = 1;
#endif
   if (nodeislocal(node)) {
	/* TODO: Get kcs port. Could check at runtime for ia64 (getenv/uname)*/
	/* Open kcs interface */
	signal(SIGALRM,sig_timeout);
	rv = ipmi_kcs_io_init(kcs_port,polltime);
	connect_state = 5;
   } else {
	struct hostent *hptr;
	char *temp;

	fprintf(fperr,"Opening connection to node %s ...\n",node);
	/* save nodename for sig_abort later */
	if (strlen(node) > 32) {
	       strncpy(nodename, node, 32); nodename[32] = 0;
	} else strcpy(nodename, node);

	/* Open lan interface */
	rv = socket(AF_INET, SOCK_DGRAM, 0); 
	if (rv < 0) return (rv);
	else sockfd = rv;

	connect_state = 1;  
	memset(&_srcaddr, 0, sizeof(_srcaddr));
	_srcaddr.sin_family = AF_INET;
	_srcaddr.sin_port = htons(0);
	_srcaddr.sin_addr.s_addr = htonl(INADDR_ANY);
	rv = bind(sockfd, (struct sockaddr *)&_srcaddr, sizeof(_srcaddr));
	if (rv < 0) return (rv);

	memset(&_destaddr, 0, sizeof(_destaddr));
	_destaddr.sin_family = AF_INET;
	_destaddr.sin_port = htons(RMCP_PRI_RMCP_PORT);  /*0x26f = 623.*/
	if ((hptr = gethostbyname(node)) == NULL) {
		printf("gnu, gethostbyname(%s): %s\n", node,hstrerror(errno));
		return(errno);
	}
	_destaddr.sin_addr = *((struct in_addr *)hptr->h_addr);
	_destaddr_len = sizeof(_destaddr);

	temp = inet_ntoa(_destaddr.sin_addr);
	strncpy(_dest_ip, temp, INET_ADDRSTRLEN);
	_dest_ip[INET_ADDRSTRLEN] = 0;

	/* Set up signals to handle errors & timeouts. */
	signal(SIGINT,sig_abort);
	signal(SIGALRM,sig_timeout);

	connect_state = 2;
	/* Send RMCP ASF ping to verify IPMI LAN connection. */
	asf_pkt[9] = 1; /*tag*/	
	rv = ipmi_lan_sendto(sockfd, asf_pkt, 12, 0,
			(struct sockaddr *)&_destaddr, _destaddr_len);
	if (fdebuglan) 
		fprintf(fperr,"ipmi_lan ping, sendto rv=%d\n",rv);
	if (rv < 0) return(GERR_CONNECT);
	connect_state = 3;

	from_len = sizeof(struct sockaddr);
	alarm(ipmi_timeout);
	rv = ipmi_lan_recvfrom(sockfd, asf_pkt, sizeof(asf_pkt), 0,
			       &from_addr,&from_len);
	alarm(0);
	if (fdebuglan) 
		fprintf(fperr,"ipmi_lan pong, recvfrom rv=%d errno=%d\n", 
			rv, errno);
	if (rv < 0) return(GERR_CONNECT);
	connect_state = 4;

	{  
		auth_type  = IPMI_SESSION_AUTH_TYPE_MD5;
		priv_level = IPMI_PRIV_LEVEL_ADMIN;
		username = user;
		authcode = pswd;
		authcode_len = (pswd) ? strlen(authcode) : 0;
	}
	alarm(ipmi_timeout);
	rv = ipmi_lan_open_session(sockfd, (struct sockaddr *)&_destaddr, 
			_destaddr_len, auth_type, username, 
			authcode, authcode_len, 0x01, priv_level, 
			&session_seqnum, &session_id);
	alarm(0);
	if (rv == 0) { /* increment seqnum if successful (session active) */
	   session_seqnum++;
	} 
	else fprintf(fperr,"ipmi_lan_open_session error, rv = %d\n",rv);
	connect_state = 5;
   }
   return(rv);
}

int ipmi_close_gnu(char *node)
{
   int rv;
   uchar buf_rs[4];  /* tmpl_cmd_set_session_priv_level_rs is 3 bytes */
   fiid_obj_t cmd_rs;

   /* use node for matching */
   if (!nodeislocal(node)) {  /* ipmi_lan, need to close & cleanup */
	if (sockfd != 0) {  /* socket is open */
	  if (session_id != 0) {  /* session is open */
  	    cmd_rs = buf_rs;
	    rv = ipmi_lan_close_session(sockfd, (struct sockaddr *)&_destaddr, 
			  _destaddr_len, auth_type, session_seqnum, session_id,
			  authcode, authcode_len, rq_seq, session_id, cmd_rs);
	  }
	  alarm(0);
	  signal(SIGALRM,SIG_DFL);
	  signal(SIGINT,SIG_DFL);
	  close(sockfd);  /*close lan socket */
	}
   } else {  /* kcs cleanup */
	alarm(0);
	signal(SIGALRM,SIG_DFL);
   }
   return (rv);
}
 
int ipmicmd_gnu(char *node, uchar cmd, uchar netfn, uchar lun, 
		uchar *pdata, uchar sdata, uchar *presp, int *sresp, 
		uchar *pcc, char fdebugcmd)
{
   fiid_obj_t cmd_rq = NULL;
   fiid_obj_t cmd_rs = NULL;
   fiid_field_t tmpl_cmd_rq[MAX_LEN_RQ+2] = {
       {8, "cmd"}, {0,""} };
   fiid_field_t tmpl_cmd_rs[MAX_LEN_RS+3] = {
       {8, "cmd"}, {8, "comp_code"}, {0,""} };
   uchar cc;
   int i,j;
   int rlen; 
   int rv = -1;

   /* check sdata/sresp against MAX_ */
   if (sdata >= MAX_LEN_RQ)  return(-4);
   if (*sresp >= MAX_LEN_RS) return(-4);
   rlen = *sresp;

   /* fill in mock tmpl_cmd_rq */
   for (i=0; i<(MAX_LEN_RQ); i++) {
	j = i+1;
	if (i < sdata) {
		tmpl_cmd_rq[j].len = 8; 
		sprintf(tmpl_cmd_rq[j].key,"data%02d",i+1);
	} else {
		tmpl_cmd_rq[j].len = 0; 
		strcpy(tmpl_cmd_rq[j].key,"");
		break;
	}
   }
   /* fill in mock tmpl_cmd_rs */
   for (i=0; i<(MAX_LEN_RS); i++) {
	j = i+2;
	if (i < rlen) {
		tmpl_cmd_rs[j].len = 8; 
		sprintf(tmpl_cmd_rs[j].key,"data%02d",i+1);
	} else {
		tmpl_cmd_rs[j].len = 0; 
		strcpy(tmpl_cmd_rs[j].key,"");
		break;
	}
   }
   
   /* Allocate cmd_rq and cmd_rs */
   // See also cmd_rq = ipmi_xcalloc (fiid_obj_len_bytes (tmpl), 1);
   if ((cmd_rq = fiid_obj_alloc(tmpl_cmd_rq)) == NULL) {
      if (fdebugcmd) 
	fprintf(fperr,"ipmicmd_gnu: fiid_obj_alloc: %s", strerror(errno));
      goto cleanup;
   }
   if ((cmd_rs = fiid_obj_alloc(tmpl_cmd_rs)) == NULL) {
      if (fdebugcmd)
	fprintf(fperr,"ipmicmd_gnu: fiid_obj_alloc: %s", strerror(errno));
      goto cleanup;
   }

   /* fill in the IPMI command and data here */
   cmd_rq[0] = cmd;
   memcpy(&cmd_rq[1],pdata,sdata);
 
   if (nodeislocal(node)) {  /*local, use kcs*/
	rv = _kcs_cmd("Gnu KCS", gnu_retry_wait_time, gnu_retry_attempts, 
		netfn, cmd_rq, tmpl_cmd_rq,
                cmd_rs, tmpl_cmd_rs,fdebugcmd);
   } else { /* ipmi_lan */
	rv = _lan_cmd(node, (struct sockaddr *)&_destaddr, _destaddr_len,
		netfn, cmd_rq, tmpl_cmd_rq,
                cmd_rs, tmpl_cmd_rs,fdebugcmd);
   }

   cc = cmd_rs[1];
   rlen = get_cmd_rslen(cmd,netfn);  /* get expected response data size */
   if (fdebugcmd) 
	fprintf(fperr,"ipmicmd_gnu: cmd = %02x rv = %d, cc = %02x, rlen = %d\n",
		cmd,rv,cc,rlen);
   if (rv == 0 && cc == 0) {  /* success */
	if (fdebugcmd) {
	   fprintf(fperr,"gnu cmd_rs(len=%d): ",rlen);
	   for (i=0; i<rlen+2; i++) fprintf(fperr,"%02x ",cmd_rs[i]);
	   fprintf(fperr,"\n");
	}
	memcpy(presp,&cmd_rs[2],rlen);
	*sresp = rlen;
   } else {  /* error */
	*sresp = 0;
	memset(presp,0,rlen);
   }
   *pcc = cc;
 
cleanup:
   fiid_obj_free(cmd_rq);
   fiid_obj_free(cmd_rs);
   return(rv);
}  /*end ipmicmd_gnu()*/

/* end ipmignu.c */
