/*
 * showsel.c
 *
 * This tool reads the firmware System Event Log records via IPMI commands.
 * It can use either the Intel /dev/imb driver or VALinux /dev/ipmikcs.
 *
 * Author:  Andy Cress  andrew.r.cress@intel.com
 *
 * Copyright (c) 2001-2002 Intel Corporation. 
 *
 * 10/16/01 Andy Cress - created
 * 10/19/01 Andy Cress - added text for Sensor Types from Table 36-3
 * 10/24/01 Andy Cress - added panic string display for OS Crit Stop
 * 11/01/01 Andy Cress - added logic to write SEL records to syslog
 * 11/15/01 Andy Cress - added logic for more description strings
 * 01/31/02 Andy Cress - isolated dependencies on /dev/imb to ipmi_cmd_ia(),
 *                       added ipmi_cmd_va() for va ipmi driver also,
 *                       added posterrs array for more descriptions.
 * 02/06/02 Andy Cress - fixed bug 279 (extra debug msg at EOF).
 * 03/25/02 Andy Cress - show free space, add -c to clear SEL.
 * 04/11/02 Andy Cress - decode timestamp into readable form
 * 06/14/02 Andy Cress - also show status error in ClearSEL
 * 07/02/02 Andy Cress v1.3 add more Usage text
 * 08/02/02 Andy Cress v1.4 moved common ipmi_cmd() code to ipmicmd.c 
 * 10/07/02 Andy Cress v1.5 added -v option with BMC version too
 * 01/16/03 Andy Cress v1.6 Handle new OS crit stop format with die code
 * 01/29/03 Andy Cress v1.7 added MV OpenIPMI support
 * 02/05/03 Andy Cress v1.8 show better message if empty SEL (cc=0xCB),
 *                          show an additional warning if free space is low.
 * 02/18/03 Andy Cress v1.9 trim out some fields so it fits on 1 line,
 *                          decode Boot Event subcodes
 * 02/27/03 Andy Cress v1.10 change OS Crit Stop decoding to handle new types
 * 03/20/03 Andy Cress v1.11 for -w, save id also, so we don't have to start
 *                           over at the beginning each time.
 * 04/30/03 Andy Cress v1.12 changed display ordering.
 * 06/23/03 Andy Cress v1.13 fix -w if log gets cleared
 * 08/19/03 Andy Cress v1.14 handle OEM/other record types
 * 09/16/03 Andy Cress v1.15 added more sens_desc strings
 * 10/03/03 Andy Cress v1.16 added more sens_desc strings (for boot)
 * 01/19/04 Andy Cress v1.17 added more sens_desc for Fans
 * 01/30/04 Andy Cress v1.18 added WIN32 flags, and sens_desc for Voltage
 * 03/12/04 Andy Cress v1.19 ClockSync description changed
 * 03/29/04 Andy Cress v1.20 change pattern matching for thresholds, 
 *                           added sens_desc for ID Button
 *                           check either time or record_id if fwritesel
 * 			     show warning if <20% free also
 *                           added sens_desc for System Events, HSC, Power, Int
 * 04/13/04 Andy Cress v1.21 added threshold OK descriptions,
 *                           change header (time is local, not GMT)
 * 05/05/04 Andy Cress v1.22 call ipmi_close before exit,
 *                           include ReportEvent code for WIN32
 * 06/10/04 Andy Cress v1.23 use gmtime instead of localtime for WIN32
 * 08/16/04 Andy Cress v1.24 added more decoding for Power events
 * 09/20/04 Andy Cress v1.25 added 2 event descriptors for ia64 platforms
 * 11/01/04 Andy Cress v1.26 add -N / -R for remote nodes   
 * 11/16/04 Andy Cress v1.27 add -U for username,
 *                           added more decoding for mBMC watchdog events
 * 11/19/04 Andy Cress v1.28 added more decoding for crit_int, slots, etc.
 *                           changed firmware error decoding.
 */
/*M*
Copyright (c) 2002, Intel Corporation
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met:

  a.. Redistributions of source code must retain the above copyright notice, 
      this list of conditions and the following disclaimer. 
  b.. Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation 
      and/or other materials provided with the distribution. 
  c.. Neither the name of Intel Corporation nor the names of its contributors 
      may be used to endorse or promote products derived from this software 
      without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR 
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON 
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *M*/
#include <stdio.h>
#ifdef WIN32
#include <windows.h>
#include "getopt.h"
#else
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <syslog.h>
#include <getopt.h>
#endif
#include <time.h>

#include "ipmicmd.h" 

#define  SELprintf  printf
 
typedef struct
{
        short              record_id;
        uchar              record_type;
        uint               timestamp;
        short              generator_id;    /*slave_addr/channel*/
        uchar              evm_rev;         //event message revision
        uchar              sensor_type;
        uchar              sensor_number;
        uchar              event_trigger;
        uchar              event_data1;
        uchar              event_data2;
        uchar              event_data3;
}       SEL_RECORD;

#define RECORD_BASE     2     //base value to the SEL record in IMB resp data
#define RID_OFFSET      0     //byte offset to the record id
#define RTYPE_OFFSET    2     //byte offset to the record type
#define RTS_OFFSET      3     //byte offset to the record timestamp
#define RGID_OFFSET     7     //byte offset to the record generator id
#define REREV_OFFSET    9     //byte offset to the record event message rev
#define RSTYPE_OFFSET   10    //byte offset to the record sensor type
#define RSN_OFFSET      11    //byte offset to the record sensor number
#define RET_OFFSET      12    //byte offset to the record event trigger
#define RDATA_OFFSET    13    //byte offset to the record event data  
#define MIN_FREE       128    /*=8*16, minimal bytes of free SEL space */
#define REC_SIZE        16    /*SEL Record Size, see IPMI 1.5 Table 26-1 */

char progname[] = "showsel";
char progver[] = "1.28";
#ifdef WIN32
char idxfile[] = "sel.idx";
HANDLE hLog = NULL;
#define SELMSG_ID  0x40000101   /* showselmsg.dll EventID 257. = 0x101 */
#else
char idxfile[] = "/usr/share/ipmiutil/sel.idx";
#endif
char fdebug = 0;
char fall = 1;
char fwritesel = 0;
char fclearsel = 0;
char fonlyver = 0;
uint savtime = 0;
short savid = 0;

static const char *sensor_types[] = {
/* 00h */ "reserved",
/* 01h */ "Temperature",
/* 02h */ "Voltage",
/* 03h */ "Current",
/* 04h */ "Fan",
/* 05h */ "Platform Chassis Intrusion",
/* 06h */ "Security Violation",
/* 07h */ "Processor",
/* 08h */ "Power Supply",
/* 09h */ "Power Unit",
/* 0Ah */ "Cooling Device",
/* 0Bh */ "FRU Sensor",
/* 0Ch */ "Memory",
/* 0Dh */ "Drive Slot",
/* 0Eh */ "POST Memory Resize",
/* 0Fh */ "System Firmware",    /*incl POST code errors*/
/* 10h */ "SEL Disabled",
/* 11h */ "Watchdog 1",
/* 12h */ "System Event",          /* offset 0,1,2 */
/* 13h */ "Critical Interrupt",        /* offset 0,1,2 */
/* 14h */ "Button",                /* offset 0,1,2 */
/* 15h */ "Board",
/* 16h */ "Microcontroller",
/* 17h */ "Add-in Card",
/* 18h */ "Chassis",
/* 19h */ "Chip Set",
/* 1Ah */ "Other FRU",
/* 1Bh */ "Cable/Interconnect",
/* 1Ch */ "Terminator",
/* 1Dh */ "System Boot Initiated",
/* 1Eh */ "Boot Error",
/* 1Fh */ "OS Boot",
/* 20h */ "OS Critical Stop",
/* 21h */ "Slot/Connector",
/* 22h */ "ACPI Power State",
/* 23h */ "Watchdog 2",
/* 24h */ "Platform Alert",
/* 25h */ "Entity Presence",
/* 26h */ "Monitor ASIC",
/* 27h */ "LAN",
/* 28h */ "Management Subsystem Health",
/* 29h */ "Battery",
};

#define NFWERRS  15
static struct {    /* See Table 36-3, type 0Fh, offset 00h */
  int code; char *msg; 
  } fwerrs[NFWERRS] = {
  0x00, "Unspecified", 
  0x01, "No system memory", 
  0x02, "No usable memory", 
  0x03, "Unrecovered Hard Disk", 
  0x04, "Unrecovered System Board", 
  0x05, "Unrecovered Diskette", 
  0x06, "Unrecovered Hard Disk Ctlr", 
  0x07, "Unrecovered PS2 USB", 
  0x08, "Boot media not found", 
  0x09, "Unrecovered video controller", 
  0x0A, "No video device", 
  0x0B, "Firmware ROM corruption", 
  0x0C, "CPU voltage mismatch", 
  0x0D, "CPU speed mismatch", 
  0x0E, "Reserved"
};

#define NFWSTAT  27
static struct {    /* See Table 36-3, type 0Fh, offset 01h & 02h */
  int code; char *msg; 
  } fwstat[NFWSTAT] = {
  0x00, "Unspecified", 
  0x01, "Memory init", 
  0x02, "Hard disk init", 
  0x03, "Secondary processor", 
  0x04, "User authentication", 
  0x05, "User-init sys setup", 
  0x06, "USB configuration", 
  0x07, "PCI configuration", 
  0x08, "Option ROM init", 
  0x09, "Video init", 
  0x0a, "Cache init", 
  0x0b, "SM Bus init", 
  0x0c, "Keyboard init", 
  0x0d, "Mgt controller", 
  0x0e, "Docking attach", 
  0x0f, "Enabling docking", 
  0x10, "Docking eject", 
  0x11, "Disabling docking", 
  0x12, "OS wake-up", 
  0x13, "Starting OS boot",
  0x14, "Baseboard init", 
  0x15, "reserved", 
  0x16, "Floppy init", 
  0x17, "Keyboard test", 
  0x18, "Mouse test", 
  0x19, "Primary processor", 
  0x1A, "Reserved"
};

#ifdef NOT
/* Remove posterrs, only applies to a specific platform's post codes, 
 * should look up POST codes in platform TPS.  - ARCress */
static struct {
  unsigned char d2;
  unsigned char d3;
  char          *msg;
  } posterrs[] = {
   0x62, 0x01, "BIOS unable to apply BIOS update to processor 1"        ,
   0x63, 0x01, "BIOS unable to apply BIOS update to processor 2"        ,
   0x64, 0x01, "BIOS does not support current stepping for processor 1" ,
   0x65, 0x01, "BIOS does not support current stepping for processor 2" ,
   0x00, 0x02, "Failure Fixed Disk"                  ,
   0x10, 0x02, "Stuck Key"                           ,
   0x11, 0x02, "Keyboard error"                      ,
   0x12, 0x02, "Keyboard Controller Failed"          ,
   0x13, 0x02, "Keyboard locked"                     ,
   0x20, 0x02, "Monitor type does not match CMOS"    ,
   0x30, 0x02, "System RAM Failed at offset"         ,
   0x31, 0x02, "Shadow RAM Failed at offset"         ,
   0x32, 0x02, "Extended RAM Failed at offset"       ,
   0x50, 0x02, "System battery is dead"              ,
   0x51, 0x02, "System CMOS checksum bad"            ,
   0x60, 0x02, "System timer error"                  ,
   0x70, 0x02, "Real time clock error"               ,
   0x97, 0x02, "ECC Memory error in bank"            ,
   0xb2, 0x02, "Incorrect Drive A: type"             ,
   0xb3, 0x02, "Incorrect Drive B: type"             ,
   0xd0, 0x02, "System cache error - Cache disabled" ,
   0xf5, 0x02, "DMA Test Failed"                     ,
   0xf6, 0x02, "Software NMI Failed"                 ,
   0x01, 0x04, "Invalid System Configuration Data" ,
   0x03, 0x04, "Resource Conflict"                 ,
   0x04, 0x04, "Resource Conflict"                 ,
   0x05, 0x04, "Expansion ROM not initialized"     ,
   0x06, 0x04, "Warning: IRQ not configured"       ,
   0x04, 0x05, "Resource Conflict"             ,
   0x05, 0x05, "Expansion ROM not initialized" ,
   0x06, 0x05, "Warning: IRQ not configured"   ,
   0x01, 0x06, "Device configuration changed"          ,
   0x02, 0x06, "Configuration error - device disabled" ,
   0x00, 0x81, "processor 0 failed BIST"                            ,
   0x01, 0x81, "processor 1 failed BIST"                            ,
   0x04, 0x81, "processor 0 Internal Error (IERR) failure"          ,
   0x05, 0x81, "processor 1 Internal Error (IERR) failure"          ,
   0x06, 0x81, "processor 0 Thermal Trip failure"                   ,
   0x07, 0x81, "processor 1 Thermal Trip failure"                   ,
   0x08, 0x81, "Watchdog Timer failed on last boot, BSP switched"   ,
   0x0a, 0x81, "processor 1 failed initialization on last boot"     ,
   0x0b, 0x81, "processor 0 failed initialization on last boot"     ,
   0x0c, 0x81, "processor 0 disabled, system in Uni-processor mode" ,
   0x0d, 0x81, "processor 1 disabled, system in Uni-processor mode" ,
   0x0e, 0x81, "processor 0 failed FRB Level 3 timer"               ,
   0x0f, 0x81, "processor 1 failed FRB Level 3 timer"               ,
   0x10, 0x81, "Server Management Interface failed to function"     ,
   0x20, 0x81, "IOP sub-system is not functional"                   ,
   0x50, 0x81, "NVRAM (CMOS) Cleared by Jumper"                     ,
   0x51, 0x81, "NVRAM (CMOS) Checksum Error, NVRAM cleared"         ,
   0x52, 0x81, "NVRAM (CMOS) Data Invalid, NVRAM cleared"           ,
   0x00, 0x00, NULL
  };

  if ((pSelRecord->event_data1==0xa0) {
	for (i=0; posterrs[i].d2 != 0x00; i++) {
		 if ((posterrs[i].d2==pSelRecord->event_data2) &&
		     (posterrs[i].d3==pSelRecord->event_data3)) {
			pstr = posterrs[i].msg;
              		break;
			}
	} /*end for*/
  }
#endif

#define NGDESC   7
struct {
 ushort g_id;
 const char desc[10];
} gen_desc[NGDESC] = {
  0x0000, "IPMB",
  0x0001, "EFI ",
  0x0003, "BIOS",
  0x0020, "BMC ",
  0x0021, "SMI ",
  0x0028, "CHAS",    /* Chassis Bridge Controller */
/*0x0031, "SMI/IPMB", * ia64/mBMC POST errors, but not in IPMI spec  */
/*0x0033, "MemCtlr",  * ia64 memory controller?  */
  0x00c0, "HSC "     /* SAF-TE Hot-Swap Controller*/
};

#define NCRITS  10
char * crit_int_str[NCRITS] = {  /* Critical Interrupt descriptions */
 /*00*/ "FP NMI  ",  /* Front Panel NMI */
 /*01*/ "Bus Timout",
 /*02*/ "IOch NMI ", /* IO channel check NMI */
 /*03*/ "Soft NMI ",
 /*04*/ "PCI PERR ",
 /*05*/ "PCI SERR ",
 /*06*/ "EISA Timout",
 /*07*/ "Bus Warn ",  /* Bus Correctable Error */
 /*08*/ "Bus Error",  /* Bus Uncorrectable Error */
 /*09*/ "Fatal NMI" };

#define NSLOTC  9
char * slot_str[NSLOTC] = {  /* Slot/Connector descriptions */
 /*00*/ "Fault   ", 
 /*01*/ "Identify",
 /*02*/ "Inserted",
 /*03*/ "InsReady",
 /*04*/ "RemReady",
 /*05*/ "PowerOff",
 /*06*/ "RemRequest",
 /*07*/ "Interlock",
 /*08*/ "Disabled",
}; 

#define NSDESC   47
struct {
 ushort genid;  /*generator id: BIOS, BMC, etc. (slave_addr/channel)*/
 uchar s_typ;   /*1=temp,2=voltage,4=fan,... */
 uchar s_num;
 uchar evtrg;   /*event trigger/type, see IPMI 1.5 table 36.1 & 36.2 */
 uchar data1;
 uchar data2;
 uchar data3;
 const char desc[40];
} sens_desc[NSDESC] = {
 0xffff,0x05, 0x05, 0x6f, 0x44, 0xff, 0xff, "LAN unplugged",    /*chassis*/
 0xffff,0x05, 0x05, 0xef, 0x44, 0xff, 0xff, "LAN restored ",    /*chassis*/
 0xffff,0x06, 0xff, 0xff, 0x45, 0xff, 0xff, "Password",    	/*security*/
 0xffff,0x07, 0xff, 0xff, 0x41, 0xff, 0xff, "Thermal trip",     /*processor*/
 0xffff,0x08, 0xff, 0x6f, 0x40, 0xff, 0xff, "Presence detected",/*power supply*/
 0xffff,0x08, 0xff, 0x6f, 0x41, 0xff, 0xff, "Failure detected", /*power supply*/
 0xffff,0x08, 0xff, 0x6f, 0x42, 0xff, 0xff, "Predictive failure",
 0xffff,0x08, 0xff, 0x6f, 0x43, 0xff, 0xff, "AC Lost",
 0xffff,0x08, 0xff, 0xef, 0x40, 0xff, 0xff, "not present",      /*power supply*/
 0xffff,0x08, 0xff, 0xef, 0x41, 0xff, 0xff, "is OK  ",          /*power supply*/
 0xffff,0x08, 0xff, 0xef, 0x42, 0xff, 0xff, "Predictive OK",
 0xffff,0x08, 0xff, 0xef, 0x43, 0xff, 0xff, "AC Regained",      /*power supply*/
 0xffff,0x09, 0xff, 0xff, 0x40, 0xff, 0xff, "Redundancy OK  ",  /*power unit*/
 0xffff,0x09, 0xff, 0xff, 0x41, 0xff, 0xff, "Redundancy Lost",
 0xffff,0x09, 0xff, 0xff, 0x44, 0xff, 0xff, "AC Lost",
 0xffff,0x09, 0xff, 0xff, 0x46, 0xff, 0xff, "Failure detected",
 0xffff,0x09, 0xff, 0xff, 0x47, 0xff, 0xff, "Predictive failure", /*power unit*/
 0xffff,0x0c, 0xff, 0xff, 0x00, 0xff, 0xff, "Correctable ECC",  /*memory*/
 0xffff,0x0f, 0x06, 0xff, 0xff, 0xff, 0xff, "POST Code",
 0xffff,0x10, 0x09, 0xff, 0x42, 0x0f, 0xff, "Log Cleared",
 0xffff,0x10, 0x09, 0xff, 0xc0, 0x03, 0xff, "ECC Memory Errors",
    /*  Often see these 3 Boot records with reboot:
     * 12 83 6f 05 00 ff = System/SEL Time Sync 1 (going down)
     * 12 83 6f 05 80 ff = System/SEL Time Sync 2 (coming up)
     * 12 83 6f 01 ff ff = OEM System Boot Event  (Completed POST) */
 0x0003,0x12, 0x83, 0xff, 0x05, 0x00, 0xff, "Boot: ClockSync 1",   
 0x0003,0x12, 0x83, 0xff, 0x05, 0x80, 0xff, "Boot: ClockSync 2",   
 0x0003,0x12, 0x83, 0xff, 0x01, 0xff, 0xff, "OEM System Booted",   
 0x0001,0x12, 0x08, 0xff, 0x01, 0xff, 0xff, "OEM System Booted",     /*ia64*/
 0x0031,0x12, 0x00, 0x6f, 0xff, 0xff, 0xff, "PEF Action",            /*ia64*/
 0x0020,0x12, 0x83, 0x6f, 0x80, 0xff, 0xff, "System Reconfigured",   /*BMC*/
 0x00c0,0x0d, 0xff, 0x08, 0x00, 0xff, 0xff, "Device Removed",    /*HSC*/
 0x00c0,0x0d, 0xff, 0x08, 0x01, 0xff, 0xff, "Device Inserted",   /*HSC*/
 0xffff,0x14, 0xff, 0xff, 0x42, 0xff, 0xff, "Reset Button pressed",
 0xffff,0x14, 0xff, 0xff, 0x40, 0xff, 0xff, "Power Button pressed",
 0xffff,0x14, 0xff, 0xff, 0x01, 0xff, 0xff, "ID Button pressed",
 0xffff,0x23, 0xff, 0xff, 0x40, 0xff, 0xff, "Expired, no action", /*watchdog2*/
 0xffff,0x23, 0xff, 0xff, 0x41, 0xff, 0xff, "Hard Reset action",  /*watchdog2*/
 0xffff,0x23, 0xff, 0xff, 0x42, 0xff, 0xff, "Power down action",  /*watchdog2*/
 0xffff,0x23, 0xff, 0xff, 0x43, 0xff, 0xff, "Power cycle action", /*watchdog2*/
 0xffff,0x23, 0xff, 0xff, 0x48, 0xff, 0xff, "Timer interrupt",    /*watchdog2*/
 0xffff,0xf3, 0x85, 0xff, 0x41, 0xff, 0xff, "State is now OK",
 0xffff,0x20, 0x00, 0xff, 0xff, 0xff, 0xff, "OS Kernel Panic",
 0xffff,0xff, 0xff, 0x01, 0x57, 0xff, 0xff, "Hi Noncrit thresh", 
 0xffff,0xff, 0xff, 0x01, 0x59, 0xff, 0xff, "Hi Crit thresh", 
 0xffff,0xff, 0xff, 0x01, 0x50, 0xff, 0xff, "Lo Noncrit thresh", 
 0xffff,0xff, 0xff, 0x01, 0x52, 0xff, 0xff, "Lo Crit thresh",
 0xffff,0xff, 0xff, 0x81, 0x57, 0xff, 0xff, "HiN thresh OK now", 
 0xffff,0xff, 0xff, 0x81, 0x59, 0xff, 0xff, "HiC thresh OK now", 
 0xffff,0xff, 0xff, 0x81, 0x50, 0xff, 0xff, "LoN thresh OK now", 
 0xffff,0xff, 0xff, 0x81, 0x52, 0xff, 0xff, "LoC thresh OK now"
		/*Thresholds apply to sensor types 1=temp, 2=voltage, 4=fan */
		/*Note that last 2 bytes usu show actual & threshold values*/
		/*evtrg: 0x01=thresh, 0x81=restored */
};


////////////////////////////////////////////////////////////////////////////////
//  GetSelEntry
////////////////////////////////////////////////////////////////////////////////
/*F*
//  Name			:	GetSelEntry
//
//  Purpose			:	This routine gets the next SEL record
//						
//  Parameters		:	
//	pRecordID		:	input/output pointer to next Record ID
//	selRecord		:	output pointer to the sel record
//
//  Returns			:	0 if success
//					-1 if failure.
//				(The error code will be in "g_iThreadLastError")
//						
//	Notes			:	uses ipmi_cmd()
//
*F*/
int GetSelEntry(short *pRecordID, SEL_RECORD *selRecord)
{
	uchar responseData[MAX_BUFFER_SIZE];
	int responseLength = MAX_BUFFER_SIZE;
	int status;
	uchar inputData[6];
	uchar completionCode;
	char* cpRecordID;	 //for next record id
	char* cpCurrentRID;  //for current record id
	char* cpCurrentTimeStamp; //for current record timestamp
	char* cpCurrentGID;  //for current record generator id
	short inRecordID;    //save the input record id

	inRecordID = *pRecordID;
	if (!selRecord) 
	{
		if (fdebug)
		    printf("GetSelEntry: error, output buffer is NULL\n");
		return -1;
	}

	//set the reservation id to zero
	inputData[0]            = 0;
	inputData[1]            = 0;

	//set the record id to get
	cpRecordID              = (char*) pRecordID;
	inputData[2]            = cpRecordID[0];
	inputData[3]            = cpRecordID[1];

	//set the offset to the record to value zero
	inputData[4]            = 0;

	//set the number of byte to read.  0xFF means read the entire record
	inputData[5]            = (char) 0xFF;

	status = ipmi_cmd(GET_SEL_ENTRY, inputData, 6, responseData, 
			&responseLength, &completionCode, fdebug);

	if (status == ACCESS_OK) {
		if( completionCode ) {
			if (completionCode == 0xCB && inRecordID == 0)
			  SELprintf("Firmware Log (SEL) is empty\n");
			else 
			  SELprintf("GetSelEntry: completion code=%x\n", 
				completionCode); // responseData[0]);
		} else {
			//successful, done
			//save the next SEL record id
			cpRecordID[0] = responseData[0];
			cpRecordID[1] = responseData[1];

			//save the SEL record content
			//save record id
			cpCurrentRID    = (char *) &(selRecord->record_id);
			cpCurrentRID[0] = responseData[RECORD_BASE+RID_OFFSET];
			cpCurrentRID[1] = responseData[RECORD_BASE+RID_OFFSET+1];

			//save record type
			selRecord->record_type = responseData[RECORD_BASE+RTYPE_OFFSET];

			//save record timestamp
			cpCurrentTimeStamp = (char *) &(selRecord->timestamp);
			cpCurrentTimeStamp[0] = responseData[RECORD_BASE+RTS_OFFSET];
			cpCurrentTimeStamp[1] = responseData[RECORD_BASE+RTS_OFFSET+1];
			cpCurrentTimeStamp[2] = responseData[RECORD_BASE+RTS_OFFSET+2];
			cpCurrentTimeStamp[3] = responseData[RECORD_BASE+RTS_OFFSET+3];

			//save record generator id
			cpCurrentGID    = (char *) &(selRecord->generator_id);
			cpCurrentGID[0] = responseData[RECORD_BASE + RGID_OFFSET];
			cpCurrentGID[1] = responseData[RECORD_BASE + RGID_OFFSET+1];

			//save record event message rev
			selRecord->evm_rev = responseData[RECORD_BASE + REREV_OFFSET];

			//save record sensor type
			selRecord->sensor_type = responseData[RECORD_BASE + RSTYPE_OFFSET];

			//save record sensor number
			selRecord->sensor_number= responseData[RECORD_BASE + RSN_OFFSET];

			//save record event trigger
			selRecord->event_trigger= responseData[RECORD_BASE + RET_OFFSET];

			//save record event data
			selRecord->event_data1	= responseData[RECORD_BASE + RDATA_OFFSET];
			selRecord->event_data2	= responseData[RECORD_BASE + RDATA_OFFSET+1];
			selRecord->event_data3	= responseData[RECORD_BASE + RDATA_OFFSET+2];

			if (inRecordID == 0 || inRecordID == selRecord->record_id)
			{
				/* We return success if the input record is 
				   begin-of-SEL (value 0), or             
                                   input and output record id matches.   
				*/
				return 0;
			}
			else 
			{
				/* If last record, inRecordID will be -1 from
				   response data, so return -1 as normal EOF. 
				   (fix to bug 279)
				*/
				if (inRecordID == -1) return(-1);
				/* If not last record, this is an error. */
				if (fdebug)
				SELprintf("GetSelEntry: input id  %d != output id %d \n", inRecordID, selRecord->record_id);
				*pRecordID = inRecordID;  //restore the input record id
				return -1;
			}
		}
	}

	// we are here because after the retry, completionCode is not COMPLETION_CODE_OK

	if (fdebug) 
		SELprintf("GetSelEntry: sendImbRequest error completion code, code=%d\n", completionCode);
 	return -1;
 
}  /* end GetSelEntry() */

void StartWriting(uint *plasttime, short *plastid)
{
	FILE *fd;
	uint lasttime; 
	short lastid;
	int ret = -1;
	
	lasttime = 0;
	lastid = 0;
	// Open the index file
	fd = fopen(idxfile,"r");
	if (fd != NULL) {
		// Read the file, get savtime & savid
		ret = fscanf(fd,"%x %x",&lasttime,&lastid);
		fclose(fd);
	}
	if (fdebug) printf("StartWriting: idx fd=%x, savtime=%x, savid=%x\n",
				fd,lasttime,(ushort)lastid);
	*plasttime = lasttime;
	*plastid = lastid;

#ifdef WIN32
	/* See Registry, under
	   HKLM/SYSTEM/CurrentControlSet/Services/EventLog/Application/showsel 
	     EventMessageFile REG_EXPAND_SZ "%SystemRoot%\system32\showsel.dll"
	     TypesSupported   REG_DWORD     0x000000007 
	*/
	hLog = RegisterEventSource(NULL, "showsel");
	if (hLog == (void *)ERROR_INVALID_HANDLE) { hLog = NULL; }
	if (hLog == NULL) 
		printf("RegisterEventSource error, %lx\n", GetLastError());
	else ret = 0; /*success*/
#else
	// Open syslog
        openlog( "SEL: ",  LOG_CONS, LOG_KERN);
	ret = 0;   /* success if here */
#endif
	if (fdebug) printf("StartWriting: ret = %d\n",ret);
	return;
}

void StopWriting(uint lasttime, short lastid)
{
	FILE *fd;
	// Rewrite the saved time & record id
	fd = fopen(idxfile,"w");
	if (fd != NULL) {
		fprintf(fd,"%x %x\n",lasttime,lastid);
		fclose(fd);
		}
#ifdef WIN32
	DeregisterEventSource(hLog);
#else
	// Close syslog
	closelog();
#endif
	return;
}

void WriteSyslog(char *msgbuf)
{
#ifdef WIN32
	boolean status;
	char *rgstrings[2] = {NULL, NULL};
	rgstrings[0] = msgbuf; /*decoded SEL entry*/
	if (hLog != NULL) {
		status = ReportEvent(hLog,EVENTLOG_INFORMATION_TYPE,
					0, SELMSG_ID, NULL, 1,0,
					rgstrings,NULL);
					/* showsel eventid = 0x101. */
		if (fdebug || (status == 0)) {  /*error or debug*/
		      printf("ReportEvent status=%d, %lx\n",
					status,GetLastError());
		} 
	}
#else
	syslog(LOG_INFO,msgbuf);
#endif
} /*end WriteSyslog*/

void ClearSEL(void)
{
	uchar responseData[MAX_BUFFER_SIZE];
	int responseLength = MAX_BUFFER_SIZE;
	int status;
	uchar inputData[6];
	uchar completionCode;
	ushort cmd;

	cmd = RESERVE_SEL;
	status = ipmi_cmd(cmd, inputData, 0, responseData, 
			&responseLength, &completionCode, fdebug);

	if (status == ACCESS_OK && completionCode == 0) {
	    cmd = CLEAR_SEL;
	    inputData[0] = responseData[0];  /*Reservation ID [0]*/
	    inputData[1] = responseData[1];  /*Reservation ID [1]*/
	    inputData[2] = 'C';
	    inputData[3] = 'L';
	    inputData[4] = 'R';
	    inputData[5] = 0xAA;   /* initiate erase */
	    status = ipmi_cmd(cmd, inputData, 6, responseData, 
	  		      &responseLength, &completionCode, fdebug);
	    /* The reservation is cancelled by the CLEAR_SEL cmd */
	    }
	if (status == ACCESS_OK) {
		if( completionCode ) {
		    SELprintf("ClearSEL: cmd = %x, completion code=%x\n",
			      cmd,completionCode); 
		} else {
		    /* Successful, done. */
		    SELprintf("ClearSEL: Log Cleared successfully\n"); 
		}
	} else {
	    SELprintf("ClearSEL: cmd = %x, error status = %x\n",
		      cmd,status); 
	}
	if (fwritesel) {  /* this won't always be enough, but try */
	   StartWriting(&savtime,&savid);
	   savid = 0;  
	   StopWriting(savtime,savid);
	}
}  /* ClearSEL()*/

char *
get_misc_desc(uchar type, uchar num, ushort genid, uchar trig,
		 uchar data1, uchar data2, uchar data3)
{
	int i;
	char *pstr = NULL; 

	/* Use sens_desc array for other misc descriptions */
	data1 &= 0x0f;  /*ignore top half of sensor offset for matching */
	for (i = 0; i < NSDESC; i++) {
	     if ((sens_desc[i].s_typ == 0xff) || 
		     (sens_desc[i].s_typ == type)) {
	      if (sens_desc[i].s_num != 0xff &&
	         sens_desc[i].s_num != num)
			    continue;
	      if (sens_desc[i].genid != 0xffff &&
	         sens_desc[i].genid != genid)
			    continue;
	      if (sens_desc[i].evtrg != 0xff &&
	         sens_desc[i].evtrg != trig)
				    continue;
	      if (sens_desc[i].data1 != 0xff && 
      	          (sens_desc[i].data1 & 0x0f) != data1)
				    continue;
	      if (sens_desc[i].data2 != 0xff && 
      	          sens_desc[i].data2 != data2)
				    continue;
	      if (sens_desc[i].data3 != 0xff && 
      	          sens_desc[i].data3 != data3)
				    continue;
	      /* have a match, use description */
	      pstr = (char *)sens_desc[i].desc; 
	      break;
 	   }
	} /*end for*/
	return(pstr);
}  /* end get_misc_desc() */

void ReadSEL(uchar mytype, char fwriteit)
{
	short RecordID = 0;  /* 0 = first record, -1 = end */
	SEL_RECORD selRecord;
	SEL_RECORD *pSelRecord = &selRecord;
	uchar dtype;
	char output[132];
	char mystr[26] = "(123)";  /*for panic string*/
	char *pstr;
	char genstr[10] = "03  ";
	char poststr[24] = "OEM Post Code = %x%x";
	char *gstr;
	int i, j;
	int rc = 0;
	time_t eventTime;
	uchar *evtime;
	uchar timebuf[40];
	/* char *e2str, *e3str; */

	memset(pSelRecord, 0, sizeof(SEL_RECORD));
	if (fwriteit) { 
		StartWriting(&savtime,&savid);
		RecordID = savid;
	}
	/* show header for the SEL records */
        printf("RecId Date/Time_______ Source_ Evt_Type SensNum Evt_detail - Trig [Evt_data]\n");
	while( rc == 0 ) {
		rc = GetSelEntry( &RecordID, pSelRecord);
		if (fwriteit && (rc != 0) && (RecordID == savid)) {
		    /* If here, log was probably cleared, so try
		     * again from the log start. */
		    RecordID = 0;
		    rc = GetSelEntry( &RecordID, pSelRecord);
		}
		if (rc != 0) { /* EOF or error */ break; }
		if (pSelRecord->record_type >= 0xe0) { /*OEM Record 26.3*/
		   char *pc;
		   pc = (char *)&pSelRecord->timestamp;  /*bytes 4:16*/
		   sprintf(output,"%04x OEM Event %02x %02x ", 
			(ushort)pSelRecord->record_id, pc[0], pc[1]);
		   j = strlen(output);
		   for (i = 2; i < 13; i++) {  /* 4:16 = 13 bytes data */
		      if (pSelRecord->record_type == 0xf0) { 
		        /* panic string will be type 0xf0 */
			if (pc[i] == 0) break;
			output[j++] = pc[i];
		      } else {
			sprintf(&output[j],"%02x ",pc[i]);
			j += 3;
		      }
		   }
		   output[j++] = '\n';
		   output[j++] = 0;
		   printf(output);
		   if (fwriteit) {
			if (pSelRecord->record_id > savid) {
			   /* Only write newer SEL records to syslog */
			   WriteSyslog(output);
			   /*savtime = pSelRecord->timestamp; No OEM timestamp*/
			   savid = pSelRecord->record_id;
			}
		   }
		} else if (pSelRecord->record_type == 0x02) {
		  /* most records are record type 2 */
		  if (mytype == 0xff || pSelRecord->sensor_type == mytype) {
		    /* set dtype, used as array index below */
		    if (pSelRecord->sensor_type > 0x29) dtype = 0;
		    else dtype = pSelRecord->sensor_type;
		    switch(pSelRecord->sensor_type) {
		     case 0x20:   /*OS Crit Stop*/
		        /* Show first 3 chars of panic string */
		        mystr[0] = '(';
		        mystr[1] = pSelRecord->sensor_number & 0x7f;
		        mystr[2] = pSelRecord->event_data2 & 0x7f;
		        mystr[3] = pSelRecord->event_data3 & 0x7f;
			mystr[4] = ')';
			mystr[5] = 0;
			pstr = mystr;
			if (pSelRecord->sensor_number & 0x80) 
				strcat(mystr,"Oops!");
			if (pSelRecord->event_data2 & 0x80) 
				strcat(mystr,"Int!");
			if (pSelRecord->event_data3 & 0x80) 
				strcat(mystr,"NullPtr!");
		        break;
		     case 0x0f:    /*Post Errs*/
			switch (pSelRecord->event_data1)
			{ 
			   case 0x00:  /* System firmware errors */
				i = pSelRecord->event_data2;
				if (i > NFWERRS) i = NFWERRS;
        		        pstr = fwerrs[i].msg;
				break;
			   case 0x01:  /* System firmware hang  */
				i = pSelRecord->event_data2;
				if (i > NFWSTAT) i = NFWSTAT;
				sprintf(poststr,"hang, %s",fwstat[i].msg);
				pstr = poststr;
				break;
			   case 0x02:  /* System firmware progress */
				i = pSelRecord->event_data2;
				if (i > NFWSTAT) i = NFWSTAT;
				sprintf(poststr,"prog, %s",fwstat[i].msg);
				pstr = poststr;
				break;
			   case 0xa0:  /* OEM post codes */
				/* OEM post codes in bytes 2 & 3 (lo-hi) */
				sprintf(poststr,"POST Code %02x%02x",
					pSelRecord->event_data3,
					pSelRecord->event_data2);
				pstr = poststr;
				break;
			   default:  
				pstr = "POST Error";  /*default string*/
			}  /*end switch(data1)*/
		        break;
		     case 0x13:   /*Crit Int*/
			i = pSelRecord->event_data1 & 0x0f;
			if (i >= NCRITS) i = NCRITS - 1;
			pstr = crit_int_str[i];
		        break;
		     case 0x21:   /*Slot/Con*/
			i = pSelRecord->event_data1 & 0x0f;
			if (i >= NSLOTC) i = NSLOTC - 1;
			sprintf(mystr,"%s",slot_str[i]);
			/* could also decode data2/data3 here if valid */
			pstr = mystr;
		        break;
		     default:   /* all other sensor types, see sens_desc */
			pstr = get_misc_desc( pSelRecord->sensor_type,
			     pSelRecord->sensor_number,
			     pSelRecord->generator_id,
			     pSelRecord->event_trigger,
      			     pSelRecord->event_data1,
      			     pSelRecord->event_data2,
      			     pSelRecord->event_data3);
			if (pstr == NULL) {
			   mystr[0] = '-'; mystr[1] = 0;
			   pstr = mystr;
			}
		    }  /*end switch(sensor_type)*/

		    /* show the Generator ID */
		    sprintf(genstr,"%04x",pSelRecord->generator_id);
		    gstr = genstr;  /* default */
		    for (i = 0; i < NGDESC; i++)
		    {
			if (gen_desc[i].g_id == pSelRecord->generator_id)
			   gstr = (char *)gen_desc[i].desc;  
		    }

		    /*firmware timestamp is #seconds since 1/1/1970 localtime*/
		    evtime = (uchar *)&pSelRecord->timestamp;
		    eventTime = evtime[0] + (evtime[1] << 8) + 
		   		(evtime[2] << 16) + (evtime[3] << 24);
		    strcpy(timebuf,"00/00/00 00:00:00");
		    strftime(timebuf,sizeof(timebuf), "%x %H:%M:%S",
#ifdef WIN32
                             gmtime(&eventTime));
#else
                             localtime(&eventTime));
#endif
		    if (fdebug) {
			uchar tbuf[40];
			char *tz;  char *lctime;
			strftime(timebuf,sizeof(timebuf), "%x %H:%M:%S %Z",
				localtime(&eventTime));
			strftime(tbuf,sizeof(tbuf), "%x %H:%M:%S %Z",
				gmtime(&eventTime));
			tz = getenv("TZ");
			if (tz == NULL) tz = "";
			lctime = getenv("LC_TIME");
			if (lctime == NULL) lctime = "";
			printf("TZ=%s, LC_TIME=%s, gmtime=%s\n",
				tz,lctime,tbuf);
		    }

		    if ((pSelRecord->event_trigger == 0x01) || /*threshold*/
		        (pSelRecord->event_trigger == 0x81)) { /*threshold ok*/
		    sprintf(output,
			"%04x %s %s %02x %s %02x %s act=%02x thr=%02x\n",
			(ushort)pSelRecord->record_id,
			timebuf,     /*  pSelRecord->timestamp */
			/* record_type,  2 = SEL record */
			/* evm_rev,   event msg rev =3 if 1.0, =4 if 1.5 */
			gstr,   /* pSelRecord->generator_id */
			pSelRecord->sensor_type,
			sensor_types[dtype] ,
			pSelRecord->sensor_number,
			pstr,
			pSelRecord->event_data2,    /*actual reading*/
			pSelRecord->event_data3 );  /*threshold value*/
		    } else {
		    sprintf(output,
			"%04x %s %s %02x %s %02x %s %02x [%02x %02x %02x]\n",
			(ushort)pSelRecord->record_id,
			timebuf,     /*  pSelRecord->timestamp */
			/* record_type,  2 = SEL record */
			/* evm_rev,   event msg rev =3 if 1.0, =4 if 1.5 */
			gstr,   /* pSelRecord->generator_id */
			pSelRecord->sensor_type,
			sensor_types[dtype] ,
			pSelRecord->sensor_number,
			pstr,
			pSelRecord->event_trigger,
			pSelRecord->event_data1,
			pSelRecord->event_data2,
			pSelRecord->event_data3 );
		    }
		    printf(output);
		    if (fwriteit) {
			/* Only write newer records */
			if ((pSelRecord->timestamp > savtime) ||
			    (pSelRecord->record_id > savid)) {
			   /* Write the given SEL record to the syslog.  */
			   WriteSyslog(output);
			   savtime = pSelRecord->timestamp;
			   savid = pSelRecord->record_id;
			}
		    } /*endif fwriteit*/
		 }  /*endif match*/
		}  /*endif type 2 */
		else {  /* other misc record type */
		   char *pc;
		   pc = (char *)&pSelRecord->record_type;
		   printf(output,"%04x Type%02x ", 
			(ushort)pSelRecord->record_id,pc[0]);
		   j = strlen(output);
		   for (i = 1; i < 14; i++) {
			sprintf(genstr,"%02x ",pc[i]);
			strcat(output,genstr);
		   }
		   strcat(output,"\n");
		   printf(output);
		   if (fwriteit) {
			/* Only write newer records  to syslog */
			if (pSelRecord->record_id > savid) {
			   WriteSyslog(output);
			   savid = pSelRecord->record_id;
			}
		   }
		}  /*endif misc type*/
		if( pSelRecord->record_id == -1 )
			break;
		if( RecordID == pSelRecord->record_id )
			break;
		memset(pSelRecord, 0, sizeof(SEL_RECORD));
	}  /*endwhile*/
	if (fwriteit) StopWriting(savtime,savid);
}  /* end ReadSEL()*/

ushort vfree = 0;
ushort vtotal = 0;

static int ReadSELinfo()
{
	uchar responseData[MAX_BUFFER_SIZE];
	int responseLength = MAX_BUFFER_SIZE;
	int status;
	uchar completionCode;
	uchar inputData[6];

	status = ipmi_cmd(GET_SEL_INFO, inputData, 0, responseData, 
			&responseLength, &completionCode, fdebug);
	if (fdebug) printf("ipmi_cmd status = %x\n",status);

	if (status == ACCESS_OK) {
		vfree = responseData[3] + (responseData[4] * 0x100);
		vtotal = responseData[1] + (responseData[2] * 0x100) + (vfree/REC_SIZE);
		SELprintf("SEL Ver %x Support %x, Size = %d records, Free space = %d records\n",
		          responseData[0],
		          responseData[13],
			  vtotal,
			  vfree/REC_SIZE );
		//successful, done
		return(0);
	} else    return(1);

}  /*end ReadSELinfo()*/

#ifdef WIN32
int __cdecl
#else
int
#endif
main(int argc, char **argv)
{
   int ret;
   char DevRecord[16];
   char c;

   printf("%s: version %s\n",progname,progver);
   while ( (c = getopt( argc, argv,"cswvxN:U:R:?")) != EOF )
      switch(c) {
          case 'x': fdebug = 1;      break;
          case 'w': fwritesel = 1;   break;
          case 's': fall = 0;        break;
          case 'c': fclearsel = 1;   break;
          case 'v': fonlyver = 1;    break;
#ifndef WIN32
	  case 'N':      /* nodename */
		gnode = strdup(optarg);
		fipmi_lan = 1;
		break;
	  case 'U':
		guser = strdup(optarg);  /*remote username */
		if (!guser)  perror("strdup");
		else {  /* Hide password from 'ps' */
	        	memset(optarg, ' ', strlen(optarg));	
		}
		break;
 	  case 'R': 
		gpswd = strdup(optarg);  /*remote password */
		if (!gpswd) perror("strdup");
		else {  /* Hide password from 'ps' */
	        	memset(optarg, ' ', strlen(optarg));	
		} 
 		break;
#endif
          default:
                printf("Usage: %s [-cswx]\n", progname);
		printf("   -c  Clears all SEL records\n");
		printf("   -s  Show only Critical Stop records\n");
		printf("   -w  Writes new SEL records to syslog\n");
		printf("   -v  Only show version information\n");
#ifndef WIN32
		printf("   -N  Nodename of target system\n");
		printf("   -U  Remote nodename's username\n");
		printf("   -R  Remote nodename's password\n");
#endif
		printf("   -x  Display extra debug messages\n");
                exit(1);
      }               
   // if (fwrite) fall = 1;   // would we ever set fwrite without fall??
#ifndef WIN32
    /* only run this as superuser */
    ret = geteuid();
    if (ret > 1) {
        printf("Not superuser (%d)\n", ret);
        exit(1);
    }
#endif

   ret = ipmi_getdeviceid( DevRecord, sizeof(DevRecord),fdebug);
   if (ret == 0) {
      uchar ipmi_maj, ipmi_min;
      ipmi_maj = DevRecord[4] & 0x0f;
      ipmi_min = DevRecord[4] >> 4;
      printf("-- BMC version %x.%x, IPMI version %d.%d \n",
             DevRecord[2],  DevRecord[3],
             ipmi_maj, ipmi_min);
   } else {
      printf("GetDeviceID error %d.  IPMI driver may not be loaded\n",ret); 
      ipmi_close();
      exit(ret);
   }

   ret = ReadSELinfo();
   if (ret == 0 && !fonlyver) {
      if (fclearsel) {
	 ClearSEL();
      } else {
	 if (fdebug) printf("%s: starting ReadSEL ...\n",progname);
         if (fall) ReadSEL(0xff,fwritesel); /* show all SEL records */
         else ReadSEL(0x20,fwritesel);      /* only show OS Crit Stops*/
	 /* PEF alerts and other log messages fail if low free space,
	    so show a warning. */
	 if (vfree < MIN_FREE) {
             printf("WARNING: free space is very low (=%d), need to clear with -c\n",
		    vfree);
	 } else if ((vfree/REC_SIZE) < ((vtotal * 20)/100)) {
             printf("WARNING: free space is low (=%d), need to clear with -c\n",
		    vfree);
	 }
      }
   }
   ipmi_close();
   exit(ret);
}

/* end showsel.c */
